import pandas as pd
from sklearn import datasets


class DatasetGenerator(object):
    @classmethod
    def generate(cls, n_samples, n_features=1, noise=10, random_state=0):
        """
        Generates a random dataset for a linear regression model
        :param n_samples: number of samples in the generated dataset
        :param n_features: the number of features that the dataset will have
        :param noise:
        :param random_state: for replication of the experiment
        :return: the samples generated for X and y, and the coefficients of the regression, all as np.arrays
        """
        X, y, coef = datasets.make_regression(
            n_samples=n_samples,
            n_features=n_features,
            n_informative=n_features,
            noise=noise,
            coef=True,
            random_state=random_state
        )
        return X, y, coef

    @classmethod
    def generate_df(cls, n_samples, features, target='y', noise=10, random_state=0):
        """
        Generates a random dataset for a linear regression model as a pandas dataframe.
        :param n_samples: number of samples in the generated dataframe
        :param features: the name of each feature as a list of str
        :param target: the name of the target (as str)
        :param noise:
        :param random_state: for replication of the experiment
        :return: the dataset generated as a pandas dataframe with the columns required in the
        params features and target and the coefficients array.
        """
        X, y, coef = cls.generate(n_samples, len(features), noise, random_state)
        X_df = pd.DataFrame(X, columns=features)
        y_df = pd.DataFrame(y, columns=[target])
        df = X_df.join(y_df)
        return df, coef

    @classmethod
    def generate_and_download(cls, filename, n_samples, features, target='y', noise=10, random_state=0):
        """
        Generates a random dataset for a linear regression model as a pandas dataframe and downloads it as csv.
        :param filename: the name of the file created as output
        :param n_samples: number of samples in the generated dataframe
        :param features: the name of each feature as a list of str
        :param target: the name of the target (as str)
        :param noise:
        :param random_state: for replication of the experiment
        :return: The coefficients array
        """
        df, coef = cls.generate_df(n_samples, features, target, noise, random_state)
        df.to_csv(filename, index=False, sep="\t")
        return coef



import numpy as np


def mean_square_error(y_pred, y):
    """
    Calculates and returns the MSE (Mean Squared Error) between 2 vectors, the test target and the prediction.
    :param y_pred: the vector of predictions
    :param y: the target
    :return: the MSE between y and y_pred
    """
    return np.mean((y - y_pred) ** 2)


def pred_diff(y_pred, y):
    """
    Returns the difference between the test target and the predicted one, for a future mse calculation.
    :param y_pred: the vector of predictions
    :param y: the target
    :return: the vector difference between y and y_pred
    """
    return y - y_pred


def sum_collection(x, y):
    """
    Returns the sum of two different vectors of the same length.
    If any of the vectors is empty, it will raise an exception.
    :param x: a vector
    :param y: another vector
    :return: The sum of the two vectors.
    """
    if len(x) != len(y):
        raise ValueError('Encrypted vectors must have the same size')
    return x + y


def deserialize(collection, encryption_service, public_key):
    """
    Deserializes a collection if the encryption service is on.
    Returns the collection as a numpy array.
    :param collection: a collection of numbers or serialized encrypted numbers if the encryption_service is on.
    :param encryption_service: the encryption service.
    :param public_key: the public key for deserializing numbers.
    :return: A numpy array with the content deserialized if the encryption service is on.
    """
    encryption_service.set_public_key(public_key)
    collection = encryption_service.get_deserialized_collection(collection) if encryption_service.is_active else collection
    return np.asarray(collection)


def serialize(collection, encryption_service, public_key):
    """
    Serializes a collection if the encryption service is on.
    Returns the collection as a list.
    :param collection: a collection of numbers or encrypted numbers if the encryption_service is on.
    :param encryption_service: the encryption service.
    :param public_key: the public key for serializing numbers.
    :return: A list with the content serialized if the encryption service is on.
    """
    encryption_service.set_public_key(public_key)
    collection = collection.tolist()
    return encryption_service.get_serialized_collection(collection) if encryption_service.is_active else collection



from commons.operations_utils.functions import deserialize, serialize

def optimized_collection_parameter(optimization, active=False):
    def wrap(f):
        def wrapped_optimized_collection_parameter(*args):
            params = args[0], optimization(args[1]) if active else args[1]
            return f(*params)
        return wrapped_optimized_collection_parameter
    return wrap


def data_owner_computation():
    def wrap(f):
        def wrapped_data_owner_computation(*args):
            encryption_service = args[0].encryption_service
            collection = args[2]
            public_key = args[3]
            collection = deserialize(collection, encryption_service, public_key)
            params = list(args)
            params[2] = collection
            result = f(*params)
            return serialize(result, encryption_service, public_key)
        return wrapped_data_owner_computation
    return wrap


def optimized_collection_response(optimization, active=False):
    def wrap(f):
        def wrapped_optimized_collection_response(*args):
            result = f(*args)
            return optimization(result) if active else result
        return wrapped_optimized_collection_response
    return wrap


def optimized_dict_collection_response(optimization, active=False):
    def wrap(f):
        def wrapped_optimized_collection_response(*args):
            result = f(*args)
            updates = list(map(lambda x: x['update']['weights'], result))
            owners = list(map(lambda x: x['data_owner'], result))  # TODO: Add model_id
            updates = optimization(updates) if active else updates
            return updates, owners
        return wrapped_optimized_collection_response
    return wrap

def normalize_optimized_collection_argument(active=False):
    def wrap(f):
        def wrapped_normalize_optimized_collection(*args):
            # TODO: Refactor!!!
            params = args[0], args[1], args[2].tolist() if active else args[2]
            return f(*params)
        return wrapped_normalize_optimized_collection
    return wrap


def normalize_optimized_response(active=False):
    def wrap(f):
        def wrapped_normalize_optimized_response(*args):
            result = f(*args)
            result['model'] = result['model'].tolist() if active else result['model']
            return result
        return wrapped_normalize_optimized_response
    return wrap



import os
from distutils.core import setup

version = os.getenv('DELTA_ML_COMMON_VERSION', 'EMPTY_VERSION')
setup(
    name='DeltaML-commons',  # How you named your package folder (MyLib)
    packages=['commons', 'commons.data', 'commons.decorators', 'commons.encryption', 'commons.model',
              'commons.operations_utils', 'commons.utils', 'commons.model.exceptions'],  # Chose the same as "name"
    version='0.9.2',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='Commons resources between DeltaML projects',  # Give a short description about your library
    author='DeltaML',  # Type in your name
    author_email='rojasagustin90@gmail.com',  # Type in your E-Mail
    url='https://github.com/DeltaML/commons',  # Provide either the link to your github or to your website
    download_url='https://github.com/DeltaML/commons/archive/v_01.tar.gz',  # I explain this later on
    keywords=['federated learning', 'machine learning', 'IA'],  # Keywords that define your package best
    install_requires=[  # I get to this in a second
        'numpy',
        'pandas',
        'sklearn',
        'phe'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
